


class Row:

    def __init__(self, data):
        self.__dict__ = data

    def __str__(self) -> str:
        return str(self.__dict__)

class Result(list):

    def __init__(self, table, data):                      
        self.extend([Row(dict(list(zip(table.keys, d)))) for d in data])    
         


# Copyright (c) 2014 Palantir Technologies
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Thread safe sqlite3 interface."""

__author__ = "Shawn Lee"
__email__ = "shawnl@palantir.com"
__license__ = "MIT"

import logging
try:
    import queue as Queue # module re-named in Python 3
except ImportError:
    import Queue
import sqlite3
import threading
import time
import uuid

LOGGER = logging.getLogger('sqlite3worker')


class Sqlite3Worker(threading.Thread):
    """Sqlite thread safe object.
    Example:
        from sqlite3worker import Sqlite3Worker
        sql_worker = Sqlite3Worker("/tmp/test.sqlite")
        sql_worker.execute(
            "CREATE TABLE tester (timestamp DATETIME, uuid TEXT)")
        sql_worker.execute(
            "INSERT into tester values (?, ?)", ("2010-01-01 13:00:00", "bow"))
        sql_worker.execute(
            "INSERT into tester values (?, ?)", ("2011-02-02 14:14:14", "dog"))
        sql_worker.execute("SELECT * from tester")
        sql_worker.close()
    """
    def __init__(self, file_name, max_queue_size=100, log_errors=False):
        """Automatically starts the thread.
        Args:
            file_name: The name of the file.
            max_queue_size: The max queries that will be queued.
        """
        threading.Thread.__init__(self)        
        self.log_errors = log_errors
        self.daemon = True
        self.sqlite3_conn = sqlite3.connect(
            file_name, check_same_thread=False,
            detect_types=sqlite3.PARSE_DECLTYPES)
        self.sqlite3_cursor = self.sqlite3_conn.cursor()
        self.sql_queue = Queue.Queue(maxsize=max_queue_size)
        self.results = {}
        self.max_queue_size = max_queue_size
        self.exit_set = False
        # Token that is put into queue when close() is called.
        self.exit_token = str(uuid.uuid4())
        self.start()
        self.thread_running = True

    def run(self):
        """Thread loop.
        This is an infinite loop.  The iter method calls self.sql_queue.get()
        which blocks if there are not values in the queue.  As soon as values
        are placed into the queue the process will continue.
        If many executes happen at once it will churn through them all before
        calling commit() to speed things up by reducing the number of times
        commit is called.
        """
        LOGGER.debug("run: Thread started")
        execute_count = 0
        for token, query, values in iter(self.sql_queue.get, None):
            LOGGER.debug("sql_queue: %s", self.sql_queue.qsize())
            if token != self.exit_token:
                LOGGER.debug("run: %s", query)
                self.run_query(token, query, values)
                execute_count += 1
                # Let the executes build up a little before committing to disk
                # to speed things up.
                if (
                        self.sql_queue.empty() or
                        execute_count == self.max_queue_size):
                    LOGGER.debug("run: commit")
                    self.sqlite3_conn.commit()
                    execute_count = 0
            # Only exit if the queue is empty. Otherwise keep getting
            # through the queue until it's empty.
            if self.exit_set and self.sql_queue.empty():
                self.sqlite3_conn.commit()
                self.sqlite3_conn.close()
                self.thread_running = False
                return

    def run_query(self, token, query, values):
        """Run a query.
        Args:
            token: A uuid object of the query you want returned.
            query: A sql query with ? placeholders for values.
            values: A tuple of values to replace "?" in query.
        """
        if query.lower().strip().startswith("select"):
            try:
                self.sqlite3_cursor.execute(query, values)
                self.results[token] = self.sqlite3_cursor.fetchall()
            except sqlite3.Error as err:
                # Put the error into the output queue since a response
                # is required.
                self.results[token] = ("Query returned error: %s: %s: %s" % (query, values, err))
                if self.log_errors: LOGGER.error("Query returned error: %s: %s: %s", query, values, err)
        else:
            try:
                self.sqlite3_cursor.execute(query, values)
            except sqlite3.Error as err:
                pass
                if self.log_errors: LOGGER.error("Query returned error: %s: %s: %s", query, values, err)

    def close(self):
        """Close down the thread and close the sqlite3 database file."""
        self.exit_set = True
        self.sql_queue.put((self.exit_token, "", ""), timeout=5)
        # Sleep and check that the thread is done before returning.
        while self.thread_running:
            time.sleep(.01)  # Don't kill the CPU waiting.

    @property
    def queue_size(self):
        """Return the queue size."""
        return self.sql_queue.qsize()

    def query_results(self, token):
        """Get the query results for a specific token.
        Args:
            token: A uuid object of the query you want returned.
        Returns:
            Return the results of the query when it's executed by the thread.
        """
        delay = .001
        while True:
            if token in self.results:
                return_val = self.results[token]
                del self.results[token]
                return return_val
            # Double back on the delay to a max of 8 seconds.  This prevents
            # a long lived select statement from trashing the CPU with this
            # infinite loop as it's waiting for the query results.
            LOGGER.debug("Sleeping: %s %s", delay, token)
            time.sleep(delay)
            if delay < 8:
                delay += delay

    def execute(self, query, values=None):
        """Execute a query.
        Args:
            query: The sql string using ? for placeholders of dynamic values.
            values: A tuple of values to be replaced into the ? of the query.
        Returns:
            If it's a select query it will return the results of the query.
        """
        if self.exit_set:
            LOGGER.debug("Exit set, not running: %s", query)
            return "Exit Called"
        LOGGER.debug("execute: %s", query)
        values = values or []
        # A token to track this query with.
        token = str(uuid.uuid4())
        # If it's a select we queue it up with a token to mark the results
        # into the output queue so we know what results are ours.
        if query.lower().strip().startswith("select"):
            self.sql_queue.put((token, query, values), timeout=5)
            return self.query_results(token)
        else:
            self.sql_queue.put((token, query, values), timeout=5)




def parse_value(value):
    if isinstance(value, str): value = f"'{value}'"
    if isinstance(value, bool): value = int(value)
    return value


class Query:

    def __init__(self):
        self.result = ""
        self.table_name = ""

    def select(self, table_name):
        self.table_name = table_name
        self.result += f"SELECT * from {table_name} "
        return self

    def update(self, table_name):
        self.table_name = table_name
        self.result += f"UPDATE {table_name} "
        return self

    def set(self, key):
        self.result += f"SET {key} "
        return self

    def delete(self, table_name):
        self.table_name = table_name
        self.result += f"DELETE from {table_name} "
        return self

    def where(self, column_name):
        if "WHERE" in self.result:
            self.result += f"AND {column_name} "
        else:    
            self.result += f"WHERE {column_name} "
        return self

    def equals(self, value):        
        self.result += f"= {parse_value(value)} "
        return self

    def greater_than(self, value):
        self.result += f"> {parse_value(value)} "
        return self
      
    def greater_or_equal(self, value):
        self.result += f">= {parse_value(value)} "
        return self

    def less_than(self, value):
        self.result += f"< {parse_value(value)} "
        return self
      
    def lesser_or_equal(self, value):
        self.result += f"<= {parse_value(value)} "
        return self


from enum import Enum, auto


class Column(Enum):
    TEXT=auto()    
    INTEGER=auto()        
    NUMERIC=auto()
    

class Table:

    def __init__(self, name):
        self.name = name
        self.columns = {}
        self.keys = []

    def add_column(self, name, type):
        if name in self.columns:
            raise RuntimeWarning(f"Column, {name}, already exists in your table!")
        self.columns[name] = type.name
        self.keys.append(name)


class Schema:

    def __init__(self, name):
        if "." in name:
            raise RuntimeWarning(
                f"You shouldn't have a period, a.k.a, '.', in your Schema name: {name}"
            )
        self.name = name
        self.tables = {}

    def add_table(self, table):
        if table.name in self.tables:
            raise RuntimeWarning(f"Table, {table.name}, already exists in your schema!")
        self.tables[table.name] = table


from setuptools import setup, find_packages


with open('README.md') as readme_file:
    README = readme_file.read()


setup_args = dict(
    name='joshsqlite',
    version='0.0.2',
    description='A thread safe DSL for sqlite3.',
    long_description_content_type="text/markdown",
    long_description=README,
    license='MIT',
    packages=find_packages(),
    author='Josh Arnold',
    author_email='dev@josharnold.me',
    keywords=['Sqlite3', 'Thread', 'Safe', 'DSL', 'Wrapper'],
    python_requires='>=3.6',
    url='https://github.com/josharnoldjosh/thread_safe_sqlite',    
)


if __name__ == '__main__':
    setup(**setup_args)


from sqlite3worker import Sqlite3Worker
from result import Result


class Database:
    """
    A thread safe sqlite3 database.
    """

    @property
    def path(self):
        """The path to the database file."""
        return self.directory+self.schema.name+".sqlite"

    def __init__(self, schema, directory="./", log_errors = False, verbose = False):        
        """Inits the database, creating a table if needed."""                
        self.schema = schema         
        if len(directory) > 0 and directory[-1] != '/': directory += '/'
        self.directory = directory
        self.log_errors = log_errors
        self.verbose = verbose
        self.db = Sqlite3Worker(self.path, log_errors=log_errors)
        self.create_tables_if_needed()

    def create_tables_if_needed(self):
        for table in self.schema.tables.values():
            try:
                columns = ", ".join([f"{key} {val}" for key, val in table.columns.items()])   
                command = f"CREATE TABLE {table.name} ({columns})"
                self.db.execute(command)            
                if self.verbose:    
                    print(f"Created '{table.name}'' with columns ({columns})")
            except Exception as e:
                if self.verbose:
                    print(e)

    def insert_into(self, table_name, **kwargs):
        """Inserts kwargs into a table."""
        if table_name not in self.schema.tables.keys():
            raise RuntimeError(f"Could not find the table, {table_name}, in database, {self.schema.name}")
        keys, values = zip(*kwargs.items())
        keys = ', '.join(keys)
        values = ', '.join(["'"+i+"'" if isinstance(i, str) else str(i) for i in values]).replace("True", "1").replace("False", "0")
        self.db.execute(f"INSERT into {table_name} ({keys}) values ({values})")

    def query(self, query):
        """Performs a given query."""
        try:
            table = self.schema.tables[query.table_name]
            data = self.db.execute(query.result)                        
            return Result(table, data)
        except Exception as e:
            if self.verbose:
                print(e)
        return []